"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from phonenumbers.phonenumberutil import NumberParseException
from phonenumbers import phonenumberutil
from phonenumbers import geocoder
from cahoots.parsers.base import BaseParser
import string


class PhoneParser(BaseParser):
    """Determines if given data is a phone number"""

    punctuation = None
    letters = None
    digits = None

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Phone", 100)

    def get_phone_number_object(self, data_string):
        """
        Takes the data_string and tries to parse a phone number out of it

        :param data_string: data that might be a phone number
        :type data_string: str
        :return: phone number data if it's a phone number
        :rtype: dict
        """
        try:
            check_region = (data_string[0] == "+")

            # First pass to see if it's a valid number
            num_obj = phonenumberutil.parse(
                data_string,
                _check_region=check_region
            )

            num_desc = geocoder.description_for_valid_number(
                num_obj, "en"
            ).strip()

        except NumberParseException:
            # If we can't parse it out, it's not a valid number
            return False

        # if we weren't able to check the region, and we didn't get a
        # description we want to modify the data and give it another go
        # with a country code added
        if not check_region and not num_desc:
            prefix = None
            if len(self.digits) == 11 and data_string[0] == "1":
                prefix = "+"
            elif len(self.digits) == 10 \
                    and (data_string[0].isdigit() or
                         data_string[0] in string.punctuation):
                prefix = "+1"

            if prefix:
                try:
                    # Second pass to see if we can get an actual
                    # geocode out of it using a hammer
                    second_pass = phonenumberutil.parse(prefix + data_string)
                    num_desc = geocoder.description_for_valid_number(
                        second_pass, "en"
                    ).strip()
                    if num_desc:
                        num_obj = second_pass
                        # confidence hit because we had to
                        # modify the data to get a result
                        self.confidence -= 5
                except (NumberParseException, Exception):
                    pass

        # Attempting to get a valid region
        num_region = phonenumberutil.region_code_for_number(num_obj)

        # This is the compiled phone number data that
        # we will use for the confidence decision
        return self.build_phone_number_dict(num_obj, num_desc, num_region)

    @classmethod
    def build_phone_number_dict(cls, num_obj, num_desc, num_region):
        """
        Builds phone number data

        :param num_obj: phone number data
        :type num_obj: phonenumbers.phonenumber.PhoneNumber
        :param num_desc: phone number description
        :type num_desc: str
        :param num_region: region code for this phone number
        :type num_region: str
        :return: phonenumber data
        :rtype: dict
        """
        return {
            'countryCode': num_obj.country_code,
            'countryCodeSource': num_obj.country_code_source,
            'nationalNumber': num_obj.national_number,
            'extension': num_obj.extension,
            'leadingZero': num_obj.italian_leading_zero,
            'carrierCode': num_obj.preferred_domestic_carrier_code,
            'description': num_desc or None,
            'region': num_region or None,
        }

    def calculate_confidence(self, data_string, phone_number_data):
        """
        Various confidence calculations

        :param data_string: the data passed to the parser
        :type data_string: string
        :param phone_number_data: data that the phone number parsing produced
        :type phone_number_data: dict
        """
        # If this is an integer, we take a big hit.
        try:
            if int(data_string) is not None:
                self.confidence -= 20
        except ValueError:
            pass

        # Length penalties
        if len(self.digits) == 10:
            # Could be a 32bit integer
            self.confidence -= 5
        if len(self.digits) < 10:
            self.confidence -= 10

        # No description == not as strong
        if not phone_number_data['description']:
            self.confidence -= 15

        # No region == not as strong
        if not phone_number_data['region']:
            self.confidence -= 15

        # if there was punctuation, and we still see
        # this as a phone number, we raise confidence
        if len(self.punctuation) > 0:
            self.confidence += (5 * len(set(self.punctuation)))

    def parse(self, data_string):
        """
        parses for phone numbers

        :param data_string: the string we want to parse
        :type data_string: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        data_string = data_string.strip()

        # 30 is a stretch...
        if len(data_string) > 30:
            return

        self.digits = [c for c in data_string if c in string.digits]

        # We're just assuming less than 7 nums = not phone
        if len(self.digits) < 7:
            return

        letter_set = set()
        self.letters = [c for c in data_string if
                        c in string.ascii_letters and
                        c not in letter_set and
                        not letter_set.add(c)]

        # We don't believe this is a phone number if more letters than nums
        if len(self.letters) > len(self.digits):
            return

        self.punctuation = [c for c in data_string if
                            c in string.punctuation or
                            c in string.whitespace]

        # Parsing our input, looking for phone numbers
        phone_number_data = self.get_phone_number_object(data_string)

        if not phone_number_data:
            return

        self.calculate_confidence(data_string, phone_number_data)

        yield self.result(
            "Phone Number",
            max(0, min(100, self.confidence)),
            phone_number_data
        )



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.parsers.programming.lexer import ProgrammingLexer
from cahoots.parsers.programming.bayesian import ProgrammingBayesianClassifier
from SereneRegistry import registry
import os
import re
import glob
import yaml


class ProgrammingParser(BaseParser):
    """Determines if a sample of text is a programming language"""

    all_keywords = []
    language_keywords = {}

    @staticmethod
    def bootstrap(config):
        """
        This method is statically called to bootstrap a parser

        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        all_keywords = []
        language_keywords = {}

        directory = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(directory, "languages/*.yaml")

        for file_path in glob.glob(path):
            with open(file_path, 'r') as language_file:
                language = yaml.load(language_file)
                all_keywords.extend(language['keywords'])
                language_keywords[language['id']] = language

        registry.set('PP_all_keywords', set(all_keywords))
        registry.set('PP_language_keywords', language_keywords)

        ProgrammingBayesianClassifier.bootstrap(config)

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Programming", 0)
        self.all_keywords = registry.get('PP_all_keywords')
        self.language_keywords = registry.get('PP_language_keywords')

    # finding common words/phrases in programming languages
    def find_common_tokens(self, dataset):
        """
        Looking for any programming language keywords in the data

        :param dataset: tokenized sample data
        :type dataset: list
        :return: tokens that match existing keywords
        :rtype: list
        """
        return [keyword for keyword in dataset if keyword in self.all_keywords]

    @classmethod
    def basic_language_heuristic(cls, language_data, dataset):
        """
        Looking for specific programming language keywords in the data

        :param language_data: the keywords for a specific programming lang
        :type language_data: dict
        :param dataset: tokenized sample data
        :type dataset: list
        :return: tokens that match existing keywords
        :rtype: list
        """
        return [keyword for keyword in
                dataset if keyword in language_data['keywords']]

    @classmethod
    def create_dataset(cls, data):
        """
        Takes a data string and converts it into a dataset for analysis

        :param data: tokenizes the potential language data
        :type data: str
        :return: tokenized data
        :rtype: list
        """
        return set(re.split('[ ;,{}()\n\t\r]', data.lower()))

    def calculate_confidence(self, lex_languages, bayes_languages):
        """
        Normalizing the parsing scores generated by lex and bayes parsing

        :param lex_languages: possible languages our lexer detected
        :type lex_languages: dict
        :param bayes_languages: possible langs our classifier detected
        :type bayes_languages: dict
        :return: the confidence scores of languages we detected
        :rtype: dict
        """
        # Adding a little bit of confidence for each lexed language we find.
        lex_scores = {}
        for lang, lex_score in lex_languages.items():
            lex_scores[lang] = lex_score / 100.0

        # Removing any scores that are less than 1% confidence
        scores = {}
        max_score = 0.0
        for lang, lex_score in lex_scores.items():
            scorecard = {
                'scores': {
                    'lexer': lex_score,
                    'bayes': bayes_languages.get(lang, 0.0)
                },
                'confidence': None
            }
            scorecard['scores']['final'] = \
                scorecard['scores']['lexer'] + scorecard['scores']['bayes']
            if scorecard['scores']['final'] > 1:
                max_score = max(max_score, scorecard['scores']['final'])
                # Rounding the result since it's a confidence value
                scores[lang] = scorecard

        # Making sure all scores are normalized to under 100
        scores = self.normalize_scores(max_score, scores)

        scores = self.prepare_scores(scores)

        return scores

    @classmethod
    def normalize_scores(cls, max_score, scores):
        """
        If the classifier scores reach over 100
        we normalize all the scores appropriately

        :param max_score: highest language score
        :type max_score: float
        :param scores: the scores for each language
        :return: dict
        """
        if max_score <= 100:
            return scores

        normal_scores = {}
        modifier = 100.0/max_score
        for lang, scorecard in scores.items():
            normal_scores[lang] = scorecard
            normal_scores[lang]['scores']['lexer'] = \
                modifier * scorecard['scores']['lexer']
            normal_scores[lang]['scores']['bayes'] = \
                modifier * scorecard['scores']['bayes']
            normal_scores[lang]['scores']['final'] = \
                modifier * scorecard['scores']['final']

        return normal_scores

    @classmethod
    def prepare_scores(cls, scores):
        """
        cleans up score values and generates integer confidence value

        :param scores: the per language scores
        :type scores: dict
        :return: the processed scores
        :rtype: dict
        """
        for _, scorecard in scores.items():
            scorecard['confidence'] = int(round(scorecard['scores']['final']))
            scorecard['scores']['lexer'] = \
                round(scorecard['scores']['lexer'], 3)
            scorecard['scores']['bayes'] = \
                round(scorecard['scores']['bayes'], 3)
            scorecard['scores']['final'] = \
                round(scorecard['scores']['final'], 3)

        return scores

    def get_possible_languages(self, dataset):
        """
        Goes through each language's keywords and finds
        languages that share keywords with the provided data set

        :param dataset: tokenized sample data
        :type dataset: list
        :return: languages that we've found tokens for
        :rtype: list
        """
        return [language for language, language_data in
                self.language_keywords.items() if
                self.basic_language_heuristic(
                    language_data, dataset
                )]

    def parse(self, data):
        """
        Determines if the data is an example of one of our trained languages

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        dataset = self.create_dataset(data)

        # Step 1: Is this possibly code?
        if not self.find_common_tokens(dataset):
            return

        # Step 2: Which languages match, based on keywords alone?
        matched_languages = self.get_possible_languages(dataset)

        # Step 3: Which languages match, based on a smarter lexer?
        lexer = ProgrammingLexer(matched_languages, data.lower())
        lex_languages = lexer.lex()

        if not lex_languages:
            return

        # Step 4: Using a Naive Bayes Classifier
        # to pinpoint the best language fits
        classifier = ProgrammingBayesianClassifier()
        bayes_languages = classifier.classify(data)

        scores = self.calculate_confidence(lex_languages, bayes_languages)

        for lang_id, scorecard in scores.items():
            yield self.result(
                self.language_keywords[lang_id]['name'],
                scorecard['confidence'],
                scorecard
            )



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from pygments.lexers.agile import PerlLexer, PythonLexer, RubyLexer
from pygments.lexers.compiled import CLexer, CppLexer
from pygments.lexers.dotnet import CSharpLexer, VbNetLexer
from pygments.lexers.jvm import JavaLexer
from pygments.lexers.web import ActionScript3Lexer, PhpLexer, JavascriptLexer
from pygments import lex
from pygments.token import Token
import threading


class ProgrammingLexerThread(threading.Thread):
    """Represents a thread that will handle one parser parsing request"""
    lexer = None
    data_string = None
    result = None

    def __init__(self, lexer_id, lexer, data_string):
        """
        :param lexer_id: the id of the lexer
        :type lexer_id: str
        :param lexer: The lexer this thread is going to use
        :type lexer: pygments.lexer.Lexer
        :param data_string: the string we're going to lex
        :type data_string: str
        """
        self.thread_id = lexer_id
        self.lexer = lexer
        self.data_string = data_string
        threading.Thread.__init__(self)

    def run(self):
        """
        Lexes the data to see what lexers can tokenize it.
        Any successful lexers are considered possible matches.
        """
        bad_tokens = (Token.Text, Token.Name, Token.Name.Other)
        tokens = [tok for tok, text in lex(self.data_string, self.lexer)
                  if tok not in bad_tokens and text != '']
        token_count = len(tokens)

        # Errors mean we definitely didn't find the right language
        if Token.Error in tokens or token_count == 0:
            self.result = False
        else:
            self.result = token_count


class ProgrammingLexer(object):
    """lexes a string with multiple programming lexers and returns tokens"""

    lexers = {
        'actionscript': ActionScript3Lexer(),
        'c': CLexer(),
        'cpp': CppLexer(),
        'cs': CSharpLexer(),
        'java': JavaLexer(),
        'javascript': JavascriptLexer(),
        'perl': PerlLexer(),
        'php': PhpLexer(startinline=True),
        'python': PythonLexer(),
        'ruby': RubyLexer(),
        'vb': VbNetLexer(),
    }

    matched_languages = []
    data = None

    def __init__(self, matched_langs, data_string):
        """
        :param matched_langs: languages that we've detected tokens for
        :type matched_langs: list
        :param data_string: the data we want to lex for possible langs
        :type data_string: str
        """
        self.matched_languages = matched_langs
        self.data = data_string

    def lex(self):
        """
        For every possible matched language, we run a lexer to see if we can
        eliminate it as a possible match. If we detect errors, or have no
        lexer matches, we remove it from the list.

        :return: the list of lexer results
        :rtype: list
        """
        results = {}
        threads = []

        # Looping through each matched language that has a lexer
        for lexer_id, lexer in \
                [[lexid, lxr] for lexid, lxr in
                 self.lexers.items() if lexid in self.matched_languages]:
            # Creating a thread for each lexer
            thread = ProgrammingLexerThread(lexer_id, lexer, self.data)
            thread.start()
            threads.append(thread)

        for thr in threads:
            thr.join()

        for thr in [th for th in threads if th.result]:
            results[thr.thread_id] = thr.result

        return results



#!/usr/bin/python
"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from flask import Flask, request
from mako.lookup import TemplateLookup
from cahoots.parser import CahootsParser
from cahootserver.config import WSGIConfig
from cahootserver import out
import os
import sys
import getopt


APP = Flask(
    __name__,
    static_folder=os.path.dirname(os.path.realpath(__file__)) + '/static'
)

# Will instantiate this later
PARSER = CahootsParser


class CahootsWSGI(object):
    """Handles all web/api responses"""

    parser = None
    template_lookup = None

    def __init__(self, cahoots_parser):
        self.parser = cahoots_parser
        self.configure_templating()

    def configure_templating(self):
        """sets up the template looker upper"""
        self.template_lookup = TemplateLookup(
            directories=self.parser.config.template['lookups'],
            module_directory=self.parser.config.template['modules']
        )

    def render(self, template, **context):
        """renders a provided template"""
        rendered_template = self.template_lookup.get_template(template)
        return rendered_template.render(**context)


class CahootsClassifier(CahootsWSGI):
    """Responsible for handling web requests"""

    def __init__(self, cahoots_parser):
        super(CahootsClassifier, self).__init__(cahoots_parser)

    def render_home(self, web_request):
        """renders web interface"""
        query = web_request.form.get('snippet', '')

        results = None
        if query != '':
            results = PARSER.parse(query)

        return self.render(
            'home.html',
            q=query,
            results=results,
            json_results=out.encode(results)
        )


class CahootsClassifierApi(CahootsWSGI):
    """Responsible for handling api requests"""

    def __init__(self, cahoots_parser):
        super(CahootsClassifierApi, self).__init__(cahoots_parser)

    @classmethod
    def render_api(cls, api_request):
        """renders cahoots response in json"""
        query = api_request.form.get('snippet', '')

        results = None
        if query != '':
            results = PARSER.parse(query)

        return out.encode(results)


@APP.route("/", methods=['POST', 'GET'])
def view_classifier():
    """displays web interface"""
    return CahootsClassifier(PARSER).render_home(request)


@APP.route("/api/", methods=['POST'])
@APP.route("/api", methods=['POST'])
def view_api():
    """displays a cahoots response in pure json"""
    return CahootsClassifierApi(PARSER).render_api(request)


def usage():
    """Cahoots Help"""
    print("")
    print("Cahoots Server Help:")
    print("")
    print("\t-h, --help")
    print("\t\tShow this help")
    print("")
    print("\t-p [port], --port [port]")
    print("\t\tSet the port the server should listen on")
    print("")
    print("\t-d, --debug")
    print("\t\tRun the server in debug mode (errors displayed, debug output)")
    print("")


def launch_server():
    """
    Runs our server!
    """
    try:
        opts, _ = getopt.getopt(
            sys.argv[1:],
            "hp:d",
            ["help", "port=", "debug"]
        )
    except getopt.GetoptError as gerror:
        print('\nError: ' + gerror.msg)
        usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            sys.exit()
        elif opt in ("-d", "--debug"):
            WSGIConfig.debug = True
        elif opt in ("-p", "--port"):
            try:
                WSGIConfig.web_port = int(arg)
                if WSGIConfig.web_port > 65535:
                    raise ValueError
            except ValueError:
                print('\nError: Invalid port')
                usage()
                sys.exit()

    # pylint: disable=global-statement
    global PARSER
    PARSER = CahootsParser(WSGIConfig, True)

    APP.run(
        host="0.0.0.0",
        port=int(os.environ.get('PORT', WSGIConfig.web_port)),
        debug=WSGIConfig.debug
    )


if __name__ == "__main__":
    launch_server()



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# pylint: disable=unnecessary-lambda
# pylint: disable=too-many-instance-attributes
from cahoots.parsers.base import BaseParser
from cahoots.parsers.equation import EquationParser
from cahoots.result import ParseResult
from binascii import unhexlify
from pyparsing import\
    Or, \
    OneOrMore, \
    Optional, \
    CaselessLiteral, \
    StringEnd, \
    ParseException, \
    replaceWith
from operator import mul
import functools
import sys
import re


class NumberParser(BaseParser):
    """determines if given data is one of many types of numbers"""

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Number", 100)

        def convert_to_literal(tok, val):
            """
            Converts a value to pyparsing caselessliteral

            :param tok: the token we want to find/replace
            :type tok: str
            :param val: the integer the token represented
            :type val: int
            :return: pyparsing caseless literal
            :rtype: pyparsing.CaselessLiteral
            """
            return CaselessLiteral(tok) \
                .setName(tok) \
                .setParseAction(replaceWith(val))

        self.defined_units = [
            ("zero", 0),
            ("oh", 0),
            ("zip", 0),
            ("zilch", 0),
            ("nada", 0),
            ("bupkis", 0),
            ("one", 1),
            ("two", 2),
            ("three", 3),
            ("four", 4),
            ("five", 5),
            ("six", 6),
            ("seven", 7),
            ("eight", 8),
            ("nine", 9),
            ("ten", 10),
            ("eleven", 11),
            ("twelve", 12),
            ("thirteen", 13),
            ("fourteen", 14),
            ("fifteen", 15),
            ("sixteen", 16),
            ("seventeen", 17),
            ("eighteen", 18),
            ("nineteen", 19),
        ]

        self.defined_tens = [
            ("ten", 10),
            ("twenty", 20),
            ("thirty", 30),
            ("forty", 40),
            ("fourty", 40),
            ("fifty", 50),
            ("sixty", 60),
            ("seventy", 70),
            ("eighty", 80),
            ("ninety", 90),
        ]

        self.defined_majors = [
            ("thousand", int(1e3)),
            ("million", int(1e6)),
            ("billion", int(1e9)),
            ("trillion", int(1e12)),
            ("quadrillion", int(1e15)),
            ("quintillion", int(1e18)),
            ("sextillion", int(1e21)),
            ("septillion", int(1e24)),
            ("octillion", int(1e27)),
            ("nonillion", int(1e30)),
            ("decillion", int(1e33)),
            ("undecillion", int(1e36)),
            ("duodecillion", int(1e39)),
            ("tredicillion", int(1e42)),
            ("quattordecillion", int(1e45)),
            ("quattuordecillion", int(1e45)),
            ("quindecillion", int(1e48)),
            ("sexdecillion", int(1e51)),
            ("septendecillion", int(1e54)),
            ("octodecillion", int(1e57)),
            ("novemdecillion", int(1e60)),
            ("vigintillion", int(1e63)),
        ]

        self.units = Or(
            [convert_to_literal(s, v) for s, v in self.defined_units]
        )
        self.tens = Or(
            [convert_to_literal(s, v) for s, v in self.defined_tens]
        )
        self.hundreds = convert_to_literal("hundred", 100)
        self.majors = Or(
            [convert_to_literal(s, v) for s, v in self.defined_majors]
        )

        self.word_product = lambda t: functools.reduce(mul, t)
        self.word_sum = lambda t: sum(t)

        self.number_partial = (
            (
                (
                    (
                        self.units + Optional(self.hundreds)
                    ).setParseAction(self.word_product) + Optional(self.tens)
                ).setParseAction(self.word_sum) ^ self.tens
            ) + Optional(self.units)).setParseAction(self.word_sum)

        self.number_words = OneOrMore(
            (self.number_partial +
             Optional(self.majors)).setParseAction(self.word_product)
        ).setParseAction(self.word_sum) + StringEnd()

        self.number_words.ignore(CaselessLiteral("-"))
        self.number_words.ignore(CaselessLiteral("and"))

        self.roman_numerals = [
            ['M', 1000],
            ['CM', 900],
            ['D', 500],
            ['CD', 400],
            ['C', 100],
            ['XC', 90],
            ['L', 50],
            ['XL', 40],
            ['X', 10],
            ['IX', 9],
            ['V', 5],
            ['IV', 4],
            ['I', 1]
        ]

    @classmethod
    def is_float(cls, data):
        """
        Checks to see if the value is a float

        :param data: potential float
        :type data: str
        :return: data confirming or denying this is a float
        :rtype: tuple
        """
        try:
            return True, float(data)
        except ValueError:
            return False, 0

    @classmethod
    def is_integer(cls, data):
        """
        Checks to see if the value is an integer

        :param data: potential integer
        :type data: str
        :return: data confirming or denying this is a integer
        :rtype: tuple
        """
        try:
            return True, int(data)
        except ValueError:
            return False, 0

    @classmethod
    def is_hex(cls, data):
        """
        Checks to see if the value is hexidecimal

        :param data: potential hexidecimal
        :type data: str
        :return: data confirming or denying this is a hexidecimal
        :rtype: tuple
        """
        if data[0] == '#':
            data = data[1:]

        try:
            return True, int(data, 16)
        except ValueError:
            return False, 0

    @classmethod
    def is_binary(cls, data):
        """
        Checks to see if the value looks like a binary

        :param data: potential binary
        :type data: str
        :return: data confirming or denying this is a binary
        :rtype: tuple
        """
        if len(data) < 4:
            return False, 0

        if len(data) % 4 != 0:
            return False, 0

        for char in data:
            if char not in ["0", "1"]:
                return False, 0

        try:
            if sys.version_info[0] < 3:  # pragma: no cover
                # pylint: disable=undefined-variable
                value = \
                    unicode(unhexlify('%x' % int(data, 2)))  # flake8: noqa
            else:  # pragma: no cover
                value = unhexlify('%x' % int(data, 2))
        except (ValueError, TypeError):
            return False, 0

        return True, value

    @classmethod
    def is_octal(cls, data):
        """
        Checks to see if the value is octal

        :param data: potential octal
        :type data: str
        :return: data confirming or denying this is a octal
        :rtype: tuple
        """
        try:
            return True, int(data, 8)
        except ValueError:
            return False, 0

    def is_roman_numeral(self, data):
        """
        Checks to see if the value is a roman numeral

        :param data: potential roman numeral
        :type data: str
        :return: data confirming or denying this is a roman numeral
        :rtype: tuple
        """
        data = data.upper()
        rgx_roman = re.compile("""^
           ([M]{0,9})   # thousands
           ([DCM]*)     # hundreds
           ([XLC]*)     # tens
           ([IVX]*)     # units
           $""", re.VERBOSE)

        match = rgx_roman.match(data)

        if match:
            value = 0
            index = 0

            while index < len(data):
                for(roman_numeral, arabic_numeral) in self.roman_numerals:
                    if data[index:(index+len(roman_numeral))] == roman_numeral:
                        value += arabic_numeral
                        index += len(roman_numeral)
                        continue

            return True, value

        return False, 0

    def is_words(self, data):
        """
        determines if the data is textual numbers

        :param data: potential textual numbers
        :type data: str
        :return: data confirming or denying this is a textual numbers
        :rtype: tuple
        """
        try:
            number_value = self.number_words.parseString(data)[0]
            return True, number_value
        except ParseException:
            return False, 0

    def is_fraction(self, data):
        """
        Detects if input is a fraction

        :param data: potential fraction
        :type data: str
        :return: data confirming or denying this is a fraction
        :rtype: tuple
        """
        if '/' not in data:
            return False, 0

        fraction_split = data.split("/")

        if len(fraction_split) > 2:
            return False, 0

        whitespace_split = data.split()
        if len(whitespace_split) > 1:
            for whitespace_section in whitespace_split:
                test = whitespace_section.strip()
                if (
                        self.is_fraction(test) == (False, 0) and
                        self.is_integer(test) == (False, 0)
                ):
                    return False, 0

        else:
            for split_section in fraction_split:
                test = split_section.strip()
                if (
                        len(test) == 0 or
                        self.is_integer(split_section.strip()) == (False, 0)
                ):
                    return False, 0

        return True, data

    @classmethod
    def prep_data(cls, data):
        """
        Removes unwanted stuff from data

        :param data: data to prep
        :type data: str
        :return: prepped data
        :rtype: str
        """
        data = data.strip()

        if not data:
            return data

        if data[0] == "-":
            data = data[1:]

        # This is a judgement call. If there's one comma, no decimal, and the
        # the comma is not in a X,XXX position, we decide it's a comma locale
        if '.' not in data and data.count(',') == 1 and data[-4:-3] != ',':
            data = data.replace(",", ".")
        else:
            data = data.replace(",", "")

        return data

    def parse_fraction(self, data):
        """
        finds and confidence scores a fraction

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        is_fraction, value = self.is_fraction(data)
        if is_fraction:
            fraction_confidence = 100

            # for every character in the data other than the /, we lower
            # the confidence a bit. Fractions are USUALLY short
            for char in data:
                if char != "/":
                    fraction_confidence -= 3

            # if the fraction isn't solve-able, we lower the confidence
            # significantly it might "technically" be a fraction made up
            # of roman numerals, etc.
            if EquationParser in self.config.enabled_modules:
                eqp = EquationParser(self.config)
                if not eqp.solve_equation(eqp.auto_float(data)):
                    fraction_confidence -= 40

            return self.result("Fraction", fraction_confidence, value)

    def parse_binary(self, data):
        """
        finds and confidence scores a binary

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        is_binary, value = self.is_binary(data)
        if is_binary:
            return self.result("Binary", 100, value)

    def parse_integer(self, data):
        """
        finds and confidence scores a integer

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        is_integer, value = self.is_integer(data)
        if is_integer:
            integer_confidence = 75
            octal_confidence = 25

            is_octal, octal_value = self.is_octal(data)
            if is_octal:
                return [
                    self.result("Integer", integer_confidence, value),
                    self.result("Octal", octal_confidence, octal_value)
                ]

            # Just an int
            return self.result("Integer", integer_confidence, value)

    def parse_float(self, data):
        """
        finds and confidence scores a float

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        if '.' in data:
            is_float, value = self.is_float(data)
            if is_float:
                return self.result("Decimal", 100, value)

    def parse_hex(self, data):
        """
        finds and confidence scores a hex number

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        if len(data) > 1:
            is_hex, value = self.is_hex(data)
            if is_hex:
                return self.result("Hexadecimal", 100, value)

    def parse_roman_numeral(self, data):
        """
        finds and confidence scores a roman numeral

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        is_roman_numeral, value = self.is_roman_numeral(data)
        if is_roman_numeral:
            return self.result("Roman Numeral", 100, value)

    def parse_word_number(self, data):
        """
        finds and confidence scores a word-number

        :param data: data we're testing
        :type data: str
        :return: results if this proper type
        :rtype: ParseResult
        """
        is_word, value = self.is_words(data)
        if is_word:
            return self.result("Word Number", 100, value)

    def parse(self, data):
        """
        parses for various types of numbers

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        data = self.prep_data(data)

        if not data:
            return

        # cascading through our potential parsers until we find something
        results = \
            self.parse_fraction(data) or \
            self.parse_binary(data) or \
            self.parse_integer(data) or \
            self.parse_float(data) or \
            self.parse_hex(data) or \
            self.parse_roman_numeral(data) or \
            self.parse_word_number(data)

        if results:
            if not isinstance(results, list):
                results = [results]
            for result in results:
                if isinstance(result, ParseResult):
                    yield result



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.parsers.location import LocationDatabase, LandmarkEntity
from cahoots.parsers.name import NameParser
from SereneRegistry import registry
import sqlite3
import string
import re


class LandmarkParser(BaseParser):
    """
    Determines if given data is a landmark

    This is dependent on if the landmark is in our database of landmarks.
    """

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Landmark", 100)

    @staticmethod
    def bootstrap(config):
        """
        This method is statically called to bootstrap a parser

        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        the_regex = re.compile('^the ', re.IGNORECASE)
        registry.set('LP_the_regex', the_regex)

    @classmethod
    def find_matching_landmarks(cls, data):
        """
        Looks in the database for landmarks matching datastring

        :param data: string we want to check against landmarks
        :type data: str
        :return: list of landmarks
        :rtype: list
        """
        database = LocationDatabase.get_database()
        cursor = database.cursor()

        try:
            rows = cursor.execute(
                'SELECT * FROM landmark WHERE resource like ?',
                (data + '%',)
            ).fetchall()
            entities = LocationDatabase.hydrate(rows, LandmarkEntity)
            entities = \
                LocationDatabase.substitute_country_data(entities, cursor)
            entities = [vars(x) for x in entities]
            database.close()
        except sqlite3.Error:
            database.close()
            return

        return entities

    def prepare_landmark_datastring(self, data):
        """
        Cleans up and validates the datastring

        :param data: data we want to check for being a location
        :type data: str
        :return: the cleaned up datastring
        :rtype: str
        """
        data = registry.get('LP_the_regex').sub('', data).strip()

        if len(data) > 75:
            return

        name_parser = NameParser(self.config)
        if not name_parser.basic_validation(data.split()):
            return

        allowed_chars = \
            string.whitespace + string.ascii_letters + string.digits
        allowed_chars += '.,-:'

        if [x for x in data if x not in allowed_chars]:
            return

        return data

    def parse(self, data):
        """
        parses for landmarks

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        data = self.prepare_landmark_datastring(data)

        if not data:
            return

        entities = self.find_matching_landmarks(data)

        if entities:
            self.confidence = \
                95-(2*len(entities)) if len(entities) > 1 else 95
            subtype = 'Single' if len(entities) == 1 else 'Multiple'
            yield self.result(subtype, self.confidence, entities)



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.result import ParseResult


class BaseParser(object):
    '''Base parser that other parsers should extend'''

    config = None
    type = "Base"
    confidence = 10

    def __init__(self, config, p_type="Base", confidence=10):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        :param p_type: parser type
        :type p_type: str
        :param confidence: result confidence default
        :type confidence: int
        """
        self.config = config
        self.type = p_type
        self.confidence = confidence

    @staticmethod
    def bootstrap(config):
        """
        This method is statically called to bootstrap a parser

        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        pass

    def parse(self, data):
        """
        Base parse method

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        raise NotImplementedError("Class must override the parse() method")

    def result(self, subtype="Unknown", confidence=0, value=None, data=None):
        """
        Returns a ParseResult object detailing the results of parsing

        :param subtype: parse result subtype
        :type subtype: str
        :param confidence: how confident we are in this result (1-100)
        :type confidence: int
        :param value: representation of the parsed data
        :type value: mixed
        :param additional_data: any additional data a parser wants to provide
        :type additional_data: mixed
        :return: parse result
        :rtype: ParseResult
        """
        if confidence == 0:
            confidence = self.confidence

        if data is None:
            data = {}

        return ParseResult(
            p_type=self.type,
            subtype=subtype,
            confidence=confidence,
            value=value,
            additional_data=data
        )



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from pyparsing import \
    Word, \
    nums, \
    CaselessLiteral, \
    Or, \
    ZeroOrMore, \
    originalTextFor, \
    Optional, \
    alphas
from cahoots.parsers.base import BaseParser
from SereneRegistry import registry
from cahoots.data import DataHandler
import os
import glob
import yaml


class MeasurementParser(BaseParser):
    """
    to be used to determine if this is an area
    described using "3 square miles" etc
    """

    @staticmethod
    def bootstrap(config):
        """
        Loads unit lists for use in this instance of the measurement parser

        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        units = {}
        systems = {}
        prepositions = DataHandler().get_prepositions()

        directory = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(directory, "units/*.yaml")

        for file_path in glob.glob(path):
            unit_file = open(file_path, 'r')
            unit_type = yaml.load(unit_file)
            for unit in unit_type['keywords']:
                units[unit] = unit_type['id']
            systems[unit_type['id']] = \
                (unit_type['system'], unit_type['type'])

        preposition_parser = \
            Or([CaselessLiteral(s) for s in prepositions]) + Word(alphas)

        measurement_parser = \
            originalTextFor(
                Word(nums, max=3) +
                ZeroOrMore(',' + Word(nums, exact=3)) +
                ZeroOrMore('.' + Word(nums)) +
                ZeroOrMore(Word(nums) + '/' + Word(nums))
            ) + \
            Or([CaselessLiteral(s) for s in units.keys()]) + \
            Optional(originalTextFor(preposition_parser))

        registry.set('MP_units', units)
        registry.set('MP_systems', systems)
        registry.set('MP_preposition_parser', preposition_parser)
        registry.set('MP_measurement_parser', measurement_parser)

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Measurement", 100)
        self.units = registry.get('MP_units')
        self.systems = registry.get('MP_systems')
        self.preposition_parser = registry.get('MP_preposition_parser')
        self.measurement_parser = registry.get('MP_measurement_parser')

    def prepare_match(self, result):
        """
        sets up value and subtype data

        :param result: the match data from pyparsing
        :type result: list
        :return: the subtype and value of the result
        :rtype: tuple
        """
        amount = \
            result.pop(0).strip() if result[0] is not None else result.pop(0)
        value = {
            'amount': amount,
            'unit': result.pop(0),
            'subject': result.pop(0) if len(result) > 0 else None,
        }

        if value['subject']:
            value['subject'] = \
                self.preposition_parser.parseString(value['subject'])[1]

        system = self.systems[self.units[value['unit']]]
        subtype = system[0] + ' ' + system[1]

        return subtype, value

    def parse(self, data):
        """
        Parses for measurements

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        if len(data) > 100:
            return

        matches = self.measurement_parser.searchString(data).asList()

        for match in matches:
            subtype, value = self.prepare_match(match)
            yield self.result(subtype, self.confidence, value)
            self.confidence -= 5

        if not matches:
            data = data.strip()
            if data in self.units:
                subtype, value = self.prepare_match([None, data])
                yield self.result(subtype, 100, value)



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# pylint: disable=eval-used
from phonenumbers.phonenumberutil import NumberParseException
from cahoots.parsers.base import BaseParser
from cahoots.util import is_number
from phonenumbers import phonenumberutil
import re
import string
# pylint: disable=unused-import
import math  # flake8: noqa


class EquationParser(BaseParser):
    """determines if given data is a mathematical equation"""

    # After we've processed the input string from the user,
    # this var will contain the assembled result
    parsed_equation = None

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Equation", 100)

    def is_simple_equation(self, data):
        """
        checking if only has symbols found in simple math equations

        :param data: potential equation text
        :type data: str
        :return: if this is a simple equation
        :rtype: bool
        """
        rgx = re.compile(r"""^
           ([()*.\-+0-9^/ ])*
           $""", re.VERBOSE)

        if rgx.match(data):
            data = self.auto_float(data)
            self.parsed_equation = self.auto_multiply(data)
            return True

        return False

    def is_text_equation(self, data):
        """
        Searching for specific textual markers that can be converted

        :param data: potential equation text
        :type data: str
        :return: if this is a text equation
        :rtype: bool
        """
        # SQUARE ROOTS
        parsed_data = re.compile(
            r'SQUARE[ ]{1,}ROOT[ ]{1,}OF[ ]{1,}\d+(\.\d+)?'
        ).sub(self.square_root_text_replace, data)

        # Simple Operators
        parsed_data = parsed_data.replace('TIMES', '*')
        parsed_data = parsed_data.replace('PLUS', '+')
        parsed_data = parsed_data.replace('MINUS', '-')
        parsed_data = parsed_data.replace('DIVIDED BY', '/')
        parsed_data = parsed_data.replace('DIVIDEDBY', '/')

        # Simple Powers
        parsed_data = re.compile(
            '[ ]{1,}SQUARED|[ ]{1,}CUBED'
        ).sub(self.simple_power_replace, parsed_data)

        if parsed_data != data:
            parsed_data = self.auto_float(parsed_data)
            self.parsed_equation = self.auto_multiply(parsed_data)
            return True

        return False

    @classmethod
    def simple_power_replace(cls, match):
        """
        Converts "SQUARED" and "CUBED" to their proper exponent form

        :param match: regex match of power text
        :type match: re.MatchObject
        :return: power text for equation
        :rtype: str
        """
        my_string = match.group()

        if my_string.find('SQUARED') != -1:
            my_string = '**2'
        elif my_string.find('CUBED') != -1:
            my_string = '**3'

        return my_string

    @classmethod
    def square_root_text_replace(cls, match):
        """
        Replaces square root references with math.sqrt

        :param match: regex match of square root text
        :type match: re.MatchObject
        :return: square root text for equation
        :rtype: str
        """
        my_string = match.group().replace('SQUARE', '')
        my_string = my_string.replace('ROOT', '')
        my_string = my_string.replace('OF', '')
        my_string = my_string.strip()

        my_string = 'math.sqrt('+my_string+')'

        return my_string

    def auto_float(self, data):
        """
        Makes all digits/decimals into floats to prevent auto-rounding

        :param data: replacing digits with float code
        :type data: str
        :return: regex pattern
        :rtype: _sre.SRE_Pattern
        """
        return re.compile(r'\d+(\.\d+)?').sub(self.float_replace, data)

    @classmethod
    def float_replace(cls, match):
        """
        This turns our numbers into floats before we eval the equation.
        This is because 4/5 comes out at 0 in python 2

        :param match: matching equation part
        :type match: re.MatchObject
        :return: prepped float string
        :rtype: str
        """
        result = 'float('+match.group()+')'
        return result

    @classmethod
    def auto_multiply(cls, data):
        """
        Any back to back parens/floats can be assumed to be
        multiplication. Adding * operator between them

        :param data: potential equation that we want to prep
        :type data: str
        :return: prepped data
        :rtype: str
        """
        data = data.replace(')float', ')*float')
        data = data.replace(')(', ')*(')

        return data

    @classmethod
    def check_for_safe_equation_string(cls, equation):
        """
        Making sure that the equation doesn't have unexpected characters
        This is pseudo-sanitization. We just make sure that the string has
        only "safe" characters. we do this by removing all expected strings,
        and seeing if we have nothing left.

        :param equation: the potential equation we want to sanitize
        :type equation: str
        :return: if the equation is clean
        :rtype: bool
        """
        # These are characters or strings that we can use in an equation
        safe_strings = \
            ['math.sqrt', 'float', '(', ')', '*', '+', '-', '/', '.']

        for safe_string in safe_strings:
            equation = equation.replace(safe_string, '')

        for num in range(10):
            equation = equation.replace(str(num), '')

        equation = equation.strip()

        return '' == equation

    def solve_equation(self, equation):
        """
        Sanitizes and Evaluates the equation to see if it's solve-able

        :param equation: potential equation
        :type equation: str
        :return: solved equation
        :type: float
        """
        if not self.check_for_safe_equation_string(equation):
            return False

        try:
            return eval(equation.strip())
        except (AttributeError, SyntaxError, TypeError):
            self.confidence = 0
            return False

    @classmethod
    def calculate_confidence(cls, data):
        """
        Calculates a confidence rating for this (possible) equation

        :param data: Data we want to get the confidence for
        :type data: str
        :return: the confidence
        :rtype: int
        """
        # We start with 100% confidence, and
        # then lower our confidence if needed.
        confidence = 100

        # lowering confidence if we have a phone number
        try:
            if len(data) <= 30 and len(data) >= 7:
                phonenumberutil.parse(data, _check_region=False)
                for _ in [c for c in data if c in string.punctuation]:
                    confidence -= 10
        except NumberParseException:
            pass

        return confidence

    @classmethod
    def clean_data(cls, data):
        """
        Removes and replaces data in prep for equation parsing

        :param data: potential equation we want to clean
        :type data: str
        :return: cleaned data
        :rtype: str
        """
        clean_data = data.upper()
        clean_data = clean_data.replace('X', '*')
        clean_data = clean_data.replace('^', '**')
        clean_data = clean_data.replace('THE', '')
        clean_data = clean_data.strip()
        return clean_data

    def parse(self, data):
        """
        Standard parse function for checking if
        entered string is a mathematical equation

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        # if we just have a number, we know this isn't an equation
        if is_number(data):
            return

        # Doing some initial data cleanup
        clean_data = self.clean_data(data)

        if not clean_data:
            return

        if self.is_simple_equation(clean_data):
            result_type = "Simple"
        elif self.is_text_equation(clean_data):
            result_type = "Text"
        else:
            return

        # If the equation proves to be solveable, we
        # calculate a confidence and report success
        calculated = self.solve_equation(self.parsed_equation)
        if calculated:
            yield self.result(
                result_type,
                self.calculate_confidence(data),
                calculated
            )



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser


class BooleanParser(BaseParser):
    '''Determines if given data is a boolean value'''

    strongTrue = ["true", "yes"]
    mediumTrue = ["yep", "yup"]
    weakTrue = ["1", "t", "one"]

    strongFalse = ["false", "no"]
    mediumFalse = ["nope"]
    weakFalse = ["0", "f", "zero"]

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "Boolean")

    @classmethod
    def is_true(cls, data):
        """
        Checks if a value is true

        :param data: potential true value
        :type data: str
        :return: if it's true or not
        :rtype: bool
        """
        if data in cls.strongTrue:
            return 100
        elif data in cls.mediumTrue:
            return 75
        elif data in cls.weakTrue:
            return 50
        return 0

    @classmethod
    def is_false(cls, data):
        """
        Checks if a value is false

        :param data: potential false value
        :type data: str
        :return: if it's false or not
        :rtype: bool
        """
        if data in cls.strongFalse:
            return 100
        elif data in cls.mediumFalse:
            return 75
        elif data in cls.weakFalse:
            return 50
        return 0

    def parse(self, data):
        """
        parses for booleans

        :param data: the string we want to parse
        :type data: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        data = data.lower()

        # The largest boolean "value" we have is 5 characters long.
        if len(data) > 5:
            return

        # Testing for true
        confidence = self.is_true(data)
        if confidence:
            yield self.result("True", confidence, True)

        # Testing for false
        confidence = self.is_false(data)
        if confidence:
            yield self.result("False", confidence, False)



# coding: utf-8
import csv
import sqlite3
import subprocess
from setuptools import setup
from setuptools.command.install import install
from distutils import log


class cahoots_install(install):
    """
    Extension of setuptools install class for post-install scripting
    """

    def prep_location_data(self):
        """Copies dist database and extracts location data"""
        location_install_cmds = [
            'cp {0}cahoots/parsers/location/data/location.sqlite.dist {0}cahoots/parsers/location/data/location.sqlite',
            'bzip2 -d -k {0}cahoots/parsers/location/data/city.txt.bz2',
            'bzip2 -d -k {0}cahoots/parsers/location/data/country.csv.bz2',
            'bzip2 -d -k {0}cahoots/parsers/location/data/street_suffix.csv.bz2',
            'bzip2 -d -k {0}cahoots/parsers/location/data/landmark.csv.bz2',
        ]

        for command in location_install_cmds:
            subprocess.call(command.format(self.install_lib), shell=True)

    def import_city_data(self, cursor):
        """Imports city data"""
        city_file = self.install_lib+'cahoots/parsers/location/data/city.txt'
        city_csv = csv.DictReader(
            open(city_file, 'r'),
            delimiter='\t',
            quotechar='"',
            fieldnames=['country', 'postal_code', 'city', 'state1', 'state2', 'province1', 'province2', 'community1', 'community2', 'latitude', 'longitude', 'coord_accuracy']
        )
        cursor.executemany(
            "INSERT INTO city (country, postal_code, city, state1, state2, province1, province2, community1, community2, latitude, longitude, coord_accuracy) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);",
            [(i['country'], i['postal_code'], i['city'], i['state1'], i['state2'], i['province1'], i['province2'], i['community1'], i['community2'], i['latitude'], i['longitude'], i['coord_accuracy'])
             for i in city_csv]
        )

    def import_country_data(self, cursor):
        """Imports country data"""
        country_file = self.install_lib+'cahoots/parsers/location/data/country.csv'
        country_csv = csv.DictReader(
            open(country_file, 'r'),
            delimiter=',',
            quotechar='"',
            fieldnames=['abbreviation', 'name']
        )
        cursor.executemany(
            "INSERT INTO country (abbreviation, name) VALUES (?, ?);",
            [(i['abbreviation'], i['name'])
             for i in country_csv]
        )

    def import_street_suffix_data(self, cursor):
        """Imports street_suffix data"""
        suffix_file = self.install_lib+'cahoots/parsers/location/data/street_suffix.csv'
        suffix_csv = csv.DictReader(
            open(suffix_file, 'r'),
            delimiter=',',
            quotechar='"',
            fieldnames=['suffix_name']
        )
        cursor.executemany(
            "INSERT INTO street_suffix (suffix_name) VALUES (?);",
            [(i['suffix_name'],)
             for i in suffix_csv]
        )

    def import_landmark_data(self, cursor):
        """Imports landmark data"""
        landmark_file = self.install_lib+'cahoots/parsers/location/data/landmark.csv'
        landmark_csv = csv.DictReader(
            open(landmark_file, 'r'),
            delimiter=',',
            quotechar='"',
            fieldnames=['resource', 'address', 'city', 'county', 'state', 'country']
        )
        cursor.executemany(
            "INSERT INTO landmark (resource, address, city, county, state, country) VALUES (?, ?, ?, ?, ?, ?);",
            [(i['resource'], i['address'], i['city'], i['county'], i['state'], i['country'])
             for i in landmark_csv]
        )

    def cleanup_location_data(self):
        """Deletes temp files"""
        location_cleanup_cmds = [
            'rm {0}cahoots/parsers/location/data/city.txt',
            'rm {0}cahoots/parsers/location/data/country.csv',
            'rm {0}cahoots/parsers/location/data/street_suffix.csv',
            'rm {0}cahoots/parsers/location/data/landmark.csv',
        ]

        for command in location_cleanup_cmds:
            subprocess.call(command.format(self.install_lib), shell=True)

    def run(self):
        """Runs the install and post-install actions"""
        install.run(self)

        log.info("Preparing/extracting location database and data for import...")
        self.prep_location_data()

        log.info("Importing location data into location database...")
        database = sqlite3.connect(self.install_lib+'cahoots/parsers/location/data/location.sqlite')
        database.text_factory = str
        cursor = database.cursor()

        log.info("Importing city data...")
        self.import_city_data(cursor)

        log.info("Importing country data...")
        self.import_country_data(cursor)

        log.info("Importing street suffix data...")
        self.import_street_suffix_data(cursor)

        log.info("Importing landmark data...")
        self.import_landmark_data(cursor)

        database.commit()
        database.close()

        log.info('Cleaning up location import temporary files...')
        self.cleanup_location_data()

        log.info('Done!')


setup (
    cmdclass = {
        'install': cahoots_install
    },
    name = 'Cahoots',
    version = '0.4.1',
    url = 'https://github.com/SerenitySoftwareLLC/cahoots',
    maintainer='Serenity Software',
    maintainer_email = 'hello@serenitysoftware.io',
    description = 'A Text Comprehension Engine in Python.',
    long_description = open('README.rst', 'r').read(),
    license = 'MIT',
    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    install_requires = [
        'watchdog',
        'flask',
        'mako',
        'dateutils',
        'pyyaml',
        'pygments',
        'pyparsing',
        'phonenumbers',
        'simplejson',
        'SereneRegistry',
        'LatLon23',
        'simplebayes',
        'validate_email',
    ],
    packages = [
        'cahoots',
        'cahoots.confidence',
        'cahoots.confidence.normalizers',
        'cahoots.parsers',
        'cahoots.parsers.location',
        'cahoots.parsers.measurement',
        'cahoots.parsers.programming',
        'cahootserver',
    ],
    package_data = {
        'cahoots': [
            # Cahoots Data
            'data/*',
            # Location
            'parsers/location/data/*.sql',
            'parsers/location/data/*.bz2',
            'parsers/location/data/location.sqlite.dist',
            'parsers/location/data/LICENSE',
            # Programming
            'parsers/programming/languages/*',
            'parsers/programming/LICENSES/*',
            'parsers/programming/trainers.zip',
            # Measurement
            'parsers/measurement/units/*',
        ],
        'cahootserver': [
            'static/*',
            'templates/*',
        ],
    },
    entry_points={
        'console_scripts': [
            'cahootserver = cahootserver.server:launch_server',
        ],
    },
)
# python ./setup.py sdist upload



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class HierarchicalNormalizerChain(object):
    """
    Orchestrates the normalization of confidence values based
    on the presence of other result types.
    """

    def __init__(self, config, types, all_types):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        :param types: list of result types
        :type types: list
        :param all_types: list of result types + subtypes
        :type all_types: list
        """
        self.config = config
        self.types = types
        self.all_types = all_types

    def normalize(self, results):
        """
        Runs all normalizers against the result set

        :param results: list of the parse result objects
        :type results: list
        :return: the parse result list, normalized
        :rtype: list
        """
        for normalizer in \
                [n for n in self.config.enabled_confidence_normalizers if
                 n.test(self.types, self.all_types)]:
            results = normalizer.normalize(results)

        # returning only results that have a condfidence greater than 0
        return [res for res in results if res.confidence > 0]



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.util import strings_intersect
import string
import socket
try:  # pragma: no cover
    # This is for Python 2 & 3 support.
    # pylint: disable=no-name-in-module
    from urllib.parse import urlparse
except ImportError:  # pragma: no cover
    # pylint: disable=import-error
    from urlparse import urlparse


class URIParser(BaseParser):
    """Determines if given data is a URI of some form"""

    def __init__(self, config):
        """
        :param config: cahoots config
        :type config: cahoots.config.BaseConfig
        """
        BaseParser.__init__(self, config, "URI", 100)

    @classmethod
    def is_ipv6_address(cls, address):
        """
        Checks if the data is an ipv6 address

        :param address: potential ip address
        :type address: str
        :return: if this is an ip addr or not
        :rtype: bool
        """
        try:
            socket.inet_pton(socket.AF_INET6, address)
            return True
        except (socket.error, UnicodeEncodeError):
            pass

        return False

    @classmethod
    def is_ipv4_address(cls, address):
        """
        checks if the data is an ipv4 address

        :param address: potential ip address
        :type address: str
        :return: if this is an ip addr or not
        :rtype: bool
        """
        try:
            socket.inet_aton(address)
            return True
        except (socket.error, UnicodeEncodeError):
            pass

        return False

    @classmethod
    def is_valid_url(cls, url):
        """
        Tries to parse a URL to see if it's valid

        :param url: text that might be a url
        :type url: str
        :return: if this is a valid url or not
        :rtype: bool
        """
        pieces = urlparse(url)

        if not all([pieces.scheme, pieces.netloc]):
            return False

        if not set(pieces.netloc) <=\
                set(string.ascii_letters + string.digits + '-.'):
            return False

        return True

    def parse(self, data_string):
        """
        parses for uris

        :param data_string: the string we want to parse
        :type data_string: str
        :return: yields parse result(s) if there are any
        :rtype: ParseResult
        """
        if len(data_string) < 4:
            return

        dot_count = data_string.count(".")

        if dot_count >= 2 or data_string.count(":") >= 2:
            if self.is_ipv4_address(data_string):
                # lowering the confidence because "Technically"
                # an ipv4 address could be a phone number
                self.confidence -= 5
                # if there's whitespace in the ip address, lower confidence
                if strings_intersect(string.whitespace, data_string):
                    self.confidence -= 50
                yield self.result("IP Address (v4)")

            elif self.is_ipv6_address(data_string):
                yield self.result("IP Address (v6)")

        letters = [c for c in data_string if c in string.ascii_letters]

        if dot_count > 0 and len(letters) >= 4:
            if self.is_valid_url(data_string):
                yield self.result("URL")

            elif '://' not in data_string:
                if self.is_valid_url('http://' + data_string):
                    # confidence hit since we had to modify the data
                    self.confidence -= 25
                    yield self.result("URL", data='http://'+data_string)
